# coding: utf-8

import os
import numpy as np
import matplotlib.pyplot as plt
import datetime
from random import choices
import random
from sklearn.mixture import GaussianMixture
import matplotlib.pyplot as plt
import time

# sample a point from P with weights W according to a cost function cost
def sample(P, W, cost):
    assert(len(P) == len(cost))
    sampled_idx = None
    if cost[0] == float('inf'):
        sampled_idx = np.random.choice(len(P))
    else:
        total_cost = np.dot(cost, W)
        if total_cost < 1e-10:
            sampled_idx = np.random.choice(len(P))
        else:   
            rand_val = np.random.sample()
            for idx in range(len(P)):
                rand_val -= (W[idx] * cost[idx]) / total_cost
                if rand_val <= 0:
                    sampled_idx = idx
                    break

    if sampled_idx is None:
        #print("No center is sampled! w=", sum(W), flush = True)
        return None
    else:
        assert(0 <= sampled_idx and sampled_idx <= len(P))
        return P[sampled_idx].copy()

# sample a point from P with Metropolis algorithm with uniform proposal distribution
# and a target distribution proportional to cost
def sample_metropolis(P, cost):
    assert(len(P) == len(cost))
    sampled_idx = 0
    if cost[0] == float('inf'):
        sampled_idx = np.random.choice(len(P))    
    else:
        for idx in range(1, len(P)):
            if cost[idx] > cost[sampled_idx]:
                sampled_idx = idx
            else:
                rand_val = np.random.sample()
                if rand_val >= cost[idx] / cost[sampled_idx]:
                    sampled_idx = idx

    if sampled_idx is None:
        print("No center is sampled!", flush = True)
        return None
    else:
        assert(0 <= sampled_idx and sampled_idx <= len(P))
        return P[sampled_idx].copy()

# computes the k-means objective of a point set P, set of centers C, weight W, 
# number of outliers Z and threshold for cost THR
def compute_objective(P, C, W=None, Z=0, THR=float('inf')):
    N = len(P)
    K = len(C)

    if W is None:
        W = [1]*N

    totalW = sum(W)
    
    obj_cost = 0
    closest = [0]*K
    best_costs = [0]*N
    best_centers = [0]*N

    for i in range(N):
        best_cost = THR
        best_center = None
        for idx in range(K):
            new_dist = P[i] - C[idx]
            new_cost = np.inner(new_dist, new_dist)
            if new_cost < best_cost:
                best_cost = new_cost
                if best_center is not None:
                    closest[best_center] -= W[i]
                best_center = idx
                closest[best_center] += W[i]
        obj_cost += W[i] * best_cost
        best_costs[i] = best_cost
        best_centers[i] = best_center

    assert(sum(closest) <= totalW)

    if Z!=0:
        pi = range(N)
        best_costs_pos = sorted([list(a) for a in zip(best_costs, pi)], key=lambda tup: tup[0])
        best_costs_sorted = [a[0] for a in best_costs_pos]
        pi = [a[1] for a in best_costs_pos]

        ZZ = Z
        for i in range(N-1, -1, -1):
            #print(best_costs_sorted[i], W[pi[i]])
            if W[pi[i]] <= ZZ:
                obj_cost -= W[pi[i]] * best_costs_sorted[i]
                if best_centers[pi[i]] != None:
                    closest[best_centers[pi[i]]] -= W[pi[i]]
                ZZ -= W[pi[i]]
            else:
                obj_cost -= ZZ * best_costs_sorted[i]
                if best_centers[pi[i]] != None:
                    closest[best_centers[pi[i]]] -= ZZ
                break;
    #assert(sum(closest) <= totalW-Z)

    return obj_cost, closest

# lloyd algorithm run on a set of points P, set of clusters C and Z outliers
# with at most max_lloyd_iter iterations
def lloyd(P, C, max_lloyd_iter=10, Z=0):
    N = len(P)
    K = len(C)
    dimension = len(P[0])

    num_lloyd_iter = 0
    converged = False
    stat = []
    while not converged:
        #print('@lloyd iteration {:d}/{:d}'.format(num_lloyd_iter + 1, max_lloyd_iter), flush = True)
        new_C = np.array([np.array([0.0]*dimension)]*K)
        C_count = [0]*K
        total_cost = 0
        cur_costs = np.zeros(N)
        best_Cs = [0]*N

        for i in range(N):
            best_cost = float('inf')
            best_C = None
            for idx in range(K):
                new_dist = P[i] - C[idx]
                new_cost = np.inner(new_dist, new_dist)
                if new_cost < best_cost:
                    best_cost = new_cost
                    best_center_idx = idx
            new_C[best_center_idx] += P[i]
            C_count[best_center_idx] += 1
            total_cost += best_cost
            cur_costs[i] = best_cost
            best_Cs[i] = best_center_idx

        for i in range(N):
            assert(np.inner(P[i] - C[best_Cs[i]], P[i] - C[best_Cs[i]]) == cur_costs[i])

        if Z > 0:
            pi = range(N)
            cur_costs_pos = sorted([list(a) for a in zip(cur_costs, pi, best_Cs)], key=lambda tup: tup[0])
            cur_costs = [a[0] for a in cur_costs_pos]
            pi = [a[1] for a in cur_costs_pos]
            best_Cs = [a[2] for a in cur_costs_pos]
            
            for i in range(N):
                assert(np.inner(P[pi[i]] - C[best_Cs[i]], P[pi[i]] - C[best_Cs[i]]) == cur_costs[i])

            C_count = [0]*K
            total_cost = 0
            new_C = np.array([np.array([0.0]*dimension)]*K)
            
            for i in range(N - Z):
                new_C[best_Cs[i]] += P[pi[i]]
                C_count[best_Cs[i]] += 1
                total_cost += cur_costs[i]
            #print("cost now is ", total_cost, " = ", compute_objective(P, C.copy(), Z=Z)[0])
            
        for idx in range(K):
            if C_count[idx] == 0: # This center is useless!
                new_C[idx] = C[idx].copy()#P[np.random.choice(len(P))]
            else:
                new_C[idx] /= C_count[idx]

        
        stat.append(total_cost)
        num_lloyd_iter += 1
        if np.all(np.equal(new_C, C)) or num_lloyd_iter >= max_lloyd_iter:
            converged = True
        else:
            C = np.array([c.copy() for c in new_C])
    obj_cost, V = compute_objective(P, C, Z=Z)
    return C, V, np.array(stat)

# k-means++
def kmeansplus(P, K, W=None, Z=0, THR=float('inf')):
    assert(W==None or len(P) == len(W))
    N = len(P)
    dimension = len(P[0])
    if W is None:
        W = [1]*N
    cost = [THR]*N
    C = []
    A = np.matrix(P)
    WT = np.matrix(W).T
    less_uniq_points_than_K = False

    for i in range(K):
        #if i * 20 // K > (i-1) * 20 // K:
        #    print('@kmeans++ sampling {:d}/{:d}'.format(i+1, K), flush = True)
        # Sample new center
        new_center = sample(P, W, cost)
        if new_center is None:
            less_uniq_points_than_K = True
            break
        C.append(new_center)

        # Update costs and candidate weights
        B = np.vstack([new_center]*N)
        new_cost = np.multiply(np.dot(np.square(A-B), np.ones((dimension,1))), WT)
        cost = [min(cost[i], new_cost[i,0]) for i in range(N)]

    obj_cost, V = compute_objective(P, C, W=W, Z=Z, THR=THR)
    assert(less_uniq_points_than_K or len(C) == K)
    assert(less_uniq_points_than_K or len(V) == K)
    statistics = (obj_cost, less_uniq_points_than_K, None)
    return C, V, statistics

# k-means++ sped up by metropolis subsampling
def kmeansplus_metropolis(P, K, W=None, Z=0, THR=float('inf'), L=10):
    N = len(P)
    dimension = len(P[0])
    if W is None:
        W = [1]*N
    cost = [THR]*L
    C = []
    less_uniq_points_than_K = False

    for i in range(K):
        # Sample new center
        sample = [random.randint(0,N-1) for idx in range(L)]
        Ps = []
        for idx in range(L):
            Ps.append(P[sample[idx]].copy())
        for idx in range(L):
            cost[idx] = THR
            for j in range(len(C)):
                cost[idx] = min(cost[idx], np.dot(Ps[idx] - C[j], Ps[idx] - C[j]))

        new_center = sample_metropolis(Ps, cost)
        if new_center is None:
            less_uniq_points_than_K = True
            break
        C.append(new_center)

        
    obj_cost, V = compute_objective(P, C, W=W, Z=Z, THR=THR)
    assert(less_uniq_points_than_K or len(C) == K)
    assert(less_uniq_points_than_K or len(V) == K)
    assert(sum(V) <= N - Z)
    statistics = (obj_cost, less_uniq_points_than_K, None)
    return C, V, statistics

# streaming k-means
def kmeansstreaming(P, K, M, Z=0, THR=float('inf')):
    N = len(P)
    K1 = 2*K
    Cs = [P[0].copy()]*(M*K1)
    Ws = [0]*(M*K1)
    
    Z1 = 0
    for m in range(M):
        C, _, _ = kmeansplus(P[m*int(N/M) : (m+1)*int(N/M)], K1, THR=THR)
        for j in range(len(C)):
            Cs[m*K1 + j] = C[j]
        for i in range(m*int(N/M), (m+1)*int(N/M)):
            cst = THR
            closest = 0
            for j in range(len(C)):
                cstj = np.dot(C[j] - P[i], C[j] - P[i])
                if cstj < cst:
                    cst = cstj
                    closest = j
            if cst == THR:
                Z1 += 1
            else:
                Ws[m*K1 + closest] += 1

    assert( sum(Ws) + Z1 == M*int(N/M))
    #print("streaming part done")
    return kmeansplus(Cs, K, W=Ws, Z=Z, THR=THR)

# random initialization (baseline)
def kmeansrand(P, K, W=None):
    N = len(P)
    dimension = len(P[0])
    if W is None:
        W = [1]*N
    
    C = []
    for i in range(K):
        C.append(P[np.random.choice(len(P))].copy())

    obj_cost, _ = compute_objective(P, C, W=W)
    statistics = (obj_cost, False, None)
    return C, False, statistics    

# sampling local search (variant of Lattanzi-Sohler algorithm) 
# can be optimized from O(nk^2) to tilde(O)(nk)
def localsearchplus(P, K, W=None, Z=0, THR=float('inf')):
    N = len(P)
    dimension = len(P[0])
    if W is None:
        W = [1]*N
    C, _, _ = kmeansplus(P, K, Z=Z, W=W, THR=THR)
    K = len(C)
    bst_C = C
    bst_obj_cost, bst_V = compute_objective(P, C, W=W, Z=Z)


    cost = [THR]*N
    for i in range(N):
        for j in range(K):
            dst = np.inner(P[i] - C[j], P[i] - C[j])
            if dst < cost[i]:
                cost[i] = dst
    
    A = np.matrix(P)
    WT = np.matrix(W).T
    less_uniq_points_than_K = False

    for it in range(K):
        
        # Sample new center
        new_center = sample(P,W, cost)
        if new_center is None:
            less_uniq_points_than_K = True
            break
        C.append(new_center)

        increase = [0] * K
        for i in range(N):
            bst_cost = float('inf')
            bst_cost_id = -1
            snd_bst_cost = float('inf')
            for j in range(K):
                dst = np.inner(P[i] - C[j], P[i] - C[j])
                if dst < bst_cost:
                    snd_bst_cost = bst_cost
                    bst_cost = dst
                    bst_cost_id = j
                elif dst < snd_bst_cost:
                    snd_bst_cost = dst
            increase[bst_cost_id] += snd_bst_cost - bst_cost
        bst_increase = float('inf')
        bstC_id = -1
        for j in range(K):
            if increase[j] < bst_increase:
                bst_increase = increase[j]
                bstC_id = j 

        C = C[0 : bstC_id] + C[bstC_id + 1 : K + 1]

        for i in range(N):
            cost[i] = THR
            for j in range(K):
                dst = np.inner(P[i] - C[j], P[i] - C[j])
                if dst < cost[i]:
                    cost[i] = dst
		
        obj_cost, V = compute_objective(P, C, W=W, Z=Z)
        if obj_cost < bst_obj_cost:
            bst_C = C
            bst_obj_cost = obj_cost
            bst_V = V

    assert(less_uniq_points_than_K or len(C) == K)
    assert(less_uniq_points_than_K or len(V) == K)
    assert(sum(V) <= N - Z)
    statistics = (bst_obj_cost, less_uniq_points_than_K, None)
    return bst_C, bst_V, statistics

# nkmeans algorithm, not used in the end
def nkmeans(P, K, Z):
    assert(Z < len(P))
    N = len(P)
    r_lo = 0.0
    r_hi = 1.0e8
    heavy = [False]*N
    
    num_iter = 0
    while(True):
        PP = []
        myZ = 0
        r = (r_lo + r_hi) / 2
        print("r = ", r)
        neighbours = [0]*N
        for i in range(N):
            for j in range(i+1, N):
                if np.dot(P[i] - P[j], P[i] - P[j]) < r:
                    neighbours[i] += 1
                    neighbours[j] += 1

            if neighbours[i] >= 2*Z:
                heavy[i] = True
                PP.append(P[i].copy())
            else:
                myZ += 1

        PP = np.array(PP)
        print(len(PP))
        if myZ < Z:
            r_hi = r
        if myZ > Z:
            r_lo = r
        if myZ == Z or num_iter >= 5:
            return kmeansplus(P=PP, K=K)                    
        num_iter += 1


# wrappers around algorithms

def run_kmeansrand(data, params):
    print('running kmeans with random init...', flush = True)
    print('data size: ', data.shape, flush = True)
    print('parameters: ', params, flush = True)
    C, V, stat = kmeansrand(P=data, K=params['K'])
    print(V)
    print('kmeans with random init terminated successfully', flush = True)
    return C, stat

def run_kmeansplus(data, params):
    print('running kmeans++...', flush = True)
    print('data size: ', data.shape, flush = True)
    print('parameters: ', params, flush = True)
    C, V, stat = kmeansplus(P=data, K=params['K'], Z=params['Z'])
    print("kmeans++ clustering: ", V)
    print('kmeans++ terminated successfully', flush = True)
    return C, stat

def run_kmeansplusstreaming(data, params):
    print('running streaming kmeans++ ...', flush = True)
    print('data size: ', data.shape, flush = True)
    print('parameters: ', params, flush = True)
    lo = 1
    hi = 1e8
    bestC, bestV, beststat = kmeansstreaming(P=data, K=params['K'], M=params['M'], Z=params['Z'], THR=float('inf'))

    for thr in np.logspace(0.0, 10.0, num = 10):
        C, V, _ = kmeansstreaming(P=data, K=params['K'], M=params['M'], Z=params['Z'], THR=thr)
        cst, _ = compute_objective(data, C, Z=Z)#we are interested in this objective
        print("for thr ", "{:.2E}".format(thr), " we have thresholded cost ", "{:.2E}".format(compute_objective(data, C, THR=thr)[0]), " and real cost ", "{:.2E}".format(compute_objective(data, C, Z=Z)[0]))
        if cst < beststat[0]:
            bestC = C
            bestV = V
            beststat = (cst, None, None)
    print("streaming kmeans++ clustering: ", V)
    print('streaming kmeans++ terminated successfully', flush = True)
    print("stat:::", beststat)
    return bestC, beststat

def run_kmeansplusthreshold(data, params):
    print('running threshold kmeans++ ...', flush = True)
    print('data size: ', data.shape, flush = True)
    print('parameters: ', params, flush = True)
    lo = 1
    hi = 1e8
    bestC, bestV, beststat = kmeansplus(P=data, K=params['K'], Z=params['Z'], THR=float('inf'))

    for thr in np.logspace(0.0, 10.0, num = 10):
        C, V, _ = kmeansplus(P=data, K=params['K'], Z=params['Z'], THR=thr)
        cst, _ = compute_objective(data, C, Z=Z)#we are interested in this objective
        print("for thr ", "{:.2E}".format(thr), " we have thresholded cost ", "{:.2E}".format(compute_objective(data, C, THR=thr)[0]), " and real cost ", "{:.2E}".format(compute_objective(data, C, Z=Z)[0]))
        if cst < beststat[0]:
            bestC = C
            bestV = V
            beststat = (cst, None, None)
    print("threshold kmeans++ clustering: ", V)
    print('threshold kmeans++ terminated successfully', flush = True)
    print("stat:::", beststat)
    return bestC, beststat

def run_kmeansplusthreshold_metropolis(data, params):
    print('running threshold kmeans++ with metropolis ...', flush = True)
    print('data size: ', data.shape, flush = True)
    print('parameters: ', params, flush = True)
    lo = 1
    hi = 1e8
    bestC, bestV, beststat = kmeansplus_metropolis(P=data, K=params['K'], Z=params['Z'], THR=float('inf'), L=params['L'])

    for thr in np.logspace(0.0, 10.0, num = 10):
        C, V, _ = kmeansplus_metropolis(P=data, K=params['K'], Z=params['Z'], THR=thr, L=params['L'])
        cst, _ = compute_objective(data, C, Z=Z)#we are interested in this objective
        print("for thr ", "{:.2E}".format(thr), " we have thresholded cost ", "{:.2E}".format(compute_objective(data, C, THR=thr)[0]), " and real cost ", "{:.2E}".format(compute_objective(data, C, Z=Z)[0]))
        if cst < beststat[0]:
            bestC = C
            bestV = V
            beststat = (cst, None, None)
    print("metropolis kmeans++ clustering: ", V)
    print('metropolis kmeans++ terminated successfully', flush = True)
    print("stat:::", beststat)
    return bestC, beststat

def run_localsearchplusthreshold(data, params):
    print('running threshold LS++ ...', flush = True)
    print('data size: ', data.shape, flush = True)
    print('parameters: ', params, flush = True)
    lo = 1
    hi = 1e8
    bestC, bestV, beststat = localsearchplus(P=data, K=params['K'], THR=float('inf'))

    for thr in np.logspace(0.0, 10.0, num = 10):
        C, V, stat = localsearchplus(P=data, K=params['K'], THR=thr)
        cst, _ = compute_objective(data, C, Z=Z)#we are interested in this objective
        print("for thr ", "{:.2E}".format(thr), " we have thresholded cost ", "{:.2E}".format(compute_objective(data, C, THR=thr)[0]), " and real cost ", "{:.2E}".format(compute_objective(data, C, Z=Z)[0]))
        if cst < beststat[0]:
            bestC = C
            bestV = V
            beststat = (cst, None, None)
    print("threshold ls++ clustering: ", V)
    print('threshold ls++ terminated successfully', flush = True)
    return bestC, beststat

def run_nkmeans(data, params):
    print('running nkmeans...', flush = True)
    print('data size: ', data.shape, flush = True)
    print('parameters: ', params, flush = True)
    C, V, stat = nkmeans(P=data, K=params['K'], Z=params['Z'])
    print("nkmeans clustering: ", V)
    print('nkmeans terminated successfully (cost below is just upper bound)', flush = True)
    return C, stat

# functions for loading the data

def load_spam_dataset(file_name):
    print('reading SPAM dataset from \'{:s}\'...'.format(file_name), flush = True)
    data = np.loadtxt(file_name, delimiter=',')
    data = np.array([tuple(x) for x in data])
    print('number of points: {:d}'.format(data.shape[0]), flush = True)
    print('dimension: {:d}'.format(data.shape[1]), flush = True)
    print('reading SPAM dataset is completed successfully.', flush = True)
    return data

def load_kddcup_dataset(data_path):
    print('reading KDDCUP dataset from \'{:s}\'...'.format(data_path), flush = True)
    lines = None
    with open(data_path, 'r') as f:
       lines = f.readlines()
    lines = [line.split(',') for line in lines]
    bad_indices = [1,2,3,len(lines[0])-1]
    filtered = []
    for line in lines:
       filtered_line = []
       for i in range(len(line)):
           if i not in bad_indices:
               filtered_line.append(float(line[i]))
       filtered.append(np.array(filtered_line))
    data = np.array(filtered)
    print('number of points: {:d}'.format(data.shape[0]), flush = True)
    print('dimension: {:d}'.format(data.shape[1]), flush = True)
    print('reading KDDCUP dataset is completed successfully.', flush = True)
    return data


# general function testing one k-means algorithm
def clustering(cl_method, params, data, data_test, num_simulations, max_lloyd_iter, Z=0):
    os.makedirs(OUTPUT_DIR, exist_ok=True)
    np.save('{:s}/params.npy'.format(OUTPUT_DIR), {'cl_method': cl_method.__name__,
        'params': params,
        'data': DATA_PATH,
        'num_simulations': num_simulations,
        'max_lloyd_iter': max_lloyd_iter})
    avg_pre_stat = 0
    avg_post_stat = 0
    avg_holdout = 0
    post_stats = []
    for i in range(num_simulations):
        print('@iteration {:d}/{:d}'.format(i+1, num_simulations), flush = True)
        pre_C, pre_stat = cl_method(data = data, params = params)
        np.save('{:s}/pre_{:02d}.npy'.format(OUTPUT_DIR, i+1), {'clusters': pre_C,
            'stat': pre_stat[0]})
        post_C, V, post_stat = lloyd(P=data, C = pre_C, max_lloyd_iter = max_lloyd_iter, Z = Z)
        np.save('{:s}/post_{:02d}.npy'.format(OUTPUT_DIR, i+1), {'clusters': post_C,
            'stat': post_stat})
        print('pre llyod cost: ', pre_stat[0], flush = True)
        print('# lloyd iter: ', len(post_stat) - 1, flush = True)
        print('post lloyd cost', post_stat, flush = True)
        avg_pre_stat += pre_stat[0]
        while(len(post_stat) != max_lloyd_iter):
            post_stat = np.append(post_stat, post_stat[-1])
        avg_post_stat += post_stat
        post_stats.append(post_stat[-1])
        #avg_holdout += compute_objective(data_test, post_C, Z=Z)[0]

    avg_pre_stat /= num_simulations
    avg_post_stat /= num_simulations
    stddev = np.std(post_stats)
    #avg_holdout /= num_simulations

    print("average pre cost: ", avg_pre_stat)
    print("average final cost: ", avg_post_stat)
    print("")
    return avg_post_stat[-1], avg_holdout, stddev

#https://archive.ics.uci.edu/ml/datasets/spambase
#DATA_PATH = '~/kmeans/spambase/spambase.data'
#http://kdd.ics.uci.edu/databases/kddcup99/kddcup99.html
#DATA_PATH = '/cluster/home/isaeed/kmeans/kddcup.data_10_percent/kddcup.data_10_percent'
DATA_PATH = 'kddcup10k.txt'
DATA_PATH_SPAM = 'spambase.data'
now = datetime.datetime.now()
OUTPUT_DIR = '{:d}{:02d}{:02d}{:02d}{:02d}{:02d}{:06d}/'.format(now.year, now.month, now.day, now.hour, now.minute, now.second, np.random.randint(1000000))

if __name__ == '__main__':
    start = time.time()
    RANDOM_STATE = 1
    m_lloyd_iter = 10
    Ks = [5*i for i in range(1, 11)]
    num_simulations = 10

    data = load_spam_dataset(DATA_PATH_SPAM)
    #data = load_kddcup_dataset(DATA_PATH)
    
    random.shuffle(data)


    data_test = []
    data = data[0 : int(len(data))]
    random.shuffle(data)
    Z = int(len(data) / 10)
 
    res_ls = []
    for K in Ks:
        res_ls.append(clustering(cl_method = run_localsearchplusthreshold,
            params = {'K': K, 'Z': Z},
            data = data,
            data_test = data_test,
            num_simulations = int(num_simulations),#'''
            max_lloyd_iter = m_lloyd_iter, 
            Z = Z))
    

    res_kmeansplus_str = []
    for K in Ks:
        res_kmeansplus_str.append(clustering(cl_method = run_kmeansplusstreaming,
            params = {'K': K, 'Z': Z, 'M': 10},
            data = data,
            data_test = data_test,
            num_simulations = int(num_simulations ),
            max_lloyd_iter = m_lloyd_iter, 
            Z = Z))


    res_kmeansplus_thr = []
    for K in Ks:
        res_kmeansplus_thr.append(clustering(cl_method = run_kmeansplusthreshold,
            params = {'K': K, 'Z': Z},
            data = data,
            data_test = data_test,
            num_simulations = int(num_simulations ),
            max_lloyd_iter = m_lloyd_iter, 
            Z = Z))

    res_kmeansplus_met = []
    for K in Ks:
        res_kmeansplus_met.append(clustering(cl_method = run_kmeansplusthreshold_metropolis,
            params = {'K': K, 'Z': Z, 'L': 100},
            data = data,
            data_test = data_test,
            num_simulations = int(num_simulations ),
            max_lloyd_iter = m_lloyd_iter, 
            Z = Z))

    
    res_kmeans = []
    for K in Ks:
        res_kmeans.append(clustering(cl_method = run_kmeansrand,
            params = {'K': K},
            data = data,
            data_test = data_test,
            num_simulations = num_simulations,
            max_lloyd_iter = m_lloyd_iter, 
            Z = Z))
    
                
    res_kmeansplus = []
    for K in Ks:
        res_kmeansplus.append(clustering(cl_method = run_kmeansplus,
            params = {'K': K, 'Z': 0},
            data = data,
            data_test = data_test,
            num_simulations = num_simulations,
            max_lloyd_iter = m_lloyd_iter, 
            Z = Z))

    legend = []
    plt.rcParams.update({'font.size': 20})
    plt.xlabel("k")
    plt.ylabel("cost")
    plt.yscale("log")
    plt.grid(True)

    data = [
        [res_kmeans, "Lloyd", "r"],
        [res_kmeansplus, "k-means++", "m"],
        [res_kmeansplus_str, "Distributed", "c"],
        [res_kmeansplus_met, "Thresholded with Metropolis", "y"],
        [res_kmeansplus_thr, "Thresholded", "b"],
        [res_ls, "Local search", "g"]
    ]

    with open(OUTPUT_DIR+"log.csv", "w") as f:
        for i in range(len(data)):
            plt.plot(
                Ks, 
                [avg for avg, bst, stddev in data[i][0]], 
                'o-'
                #[stddev for avg, bst, stddev in data[i][0]],
                #capsize = 20
                ) 
                #linestyle = data[i][2]+"-")
            legend.append(data[i][1])
            f.write(','.join([str(avg) for avg, bst, stddev in data[i][0]]) + '\n')
            f.write(','.join([str(stddev) for avg, bst, stddev in data[i][0]]) + '\n')


    plt.legend(legend)
    plt.show()
        
    end = time.time()
        
    print("time:", end - start)

    



